'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var EFFECT_ERROR = "effect/error";

/**
 * Update containing new state-data if any, and any messages to send to
 * supervisors.
 */

/**
 * Initialization function, called when the initial data is loaded into the
 * state.
 */

/**
 * Message representing any other kind of message in the type-signature of
 * `ModelUpdate`, do not match against this message directly.
 *
 * This type exists to ensure users refine the message-type before acting
 * on incoming messages.
 */

/**
 * Update function, receives messages for a state and produces a state-update
 * which can include messages sent to supervisors.
 *
 * The incoming messages can be any random message, refine on `tag` to ensure
 * the messages are indended for the state before acting upon them.
 */

/**
 * Definition of a state containing the data `T` which can be instantiated given
 * the initial data `I`.
 */

/**
 * Utility type resolving the data-type of a model with unspecified
 * type-parameters.
 */

/**
 * Utility type resolving the init-type of a model with unspecified
 * type-parameters.
 */
function updateNone() {
  return 1;
}
/**
 * Creates an update replacing the data of the state.
 */

function updateData(data) {
  return {
    data: data,
    messages: null
  };
}
/**
 * Creates an update replacing the data of the state, and sends a list o
 * messages to supervisoring states.
 */

function updateAndSend(data) {
  for (var _len = arguments.length, messages = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    messages[_key - 1] = arguments[_key];
  }

  return {
    data: data,
    messages: messages
  };
}

function _inheritsLoose(subClass, superClass) {
  subClass.prototype = Object.create(superClass.prototype);
  subClass.prototype.constructor = subClass;
  subClass.__proto__ = superClass;
}

/**
 * Debug assertion, will be a noop in production builds.
 */
var debugAssert = function debugAssert(cond, msg) {
  if (process.env.NODE_ENV !== "production" && !cond) {
    throw new Error("Assertion failed: " + msg);
  }
};

var EventEmitter =
/*#__PURE__*/
function () {
  function EventEmitter() {// Empty

    this._eventListeners = {};
  }

  var _proto = EventEmitter.prototype;

  _proto.addListener = function addListener(eventName, // Using inline Listener here since E is not allowed in an input position
  listener) {
    var existing = this._eventListeners[eventName];

    if (existing) {
      existing.push(listener);
    } else {
      this._eventListeners[eventName] = [listener];
    }
  };

  _proto.removeListener = function removeListener(eventName, listener) {
    var existing = this._eventListeners[eventName];

    if (existing) {
      var i = existing.indexOf(listener);

      if (i >= 0) {
        if (existing.length > 1) {
          existing.splice(i, 1);
        } else {
          delete this._eventListeners[eventName];
        }
      }
    }
  };

  _proto.removeAllListeners = function removeAllListeners(eventName) {
    if (eventName) {
      delete this._eventListeners[eventName];
    } else {
      this._eventListeners = {};
    }
  };

  _proto.listeners = function listeners(eventName) {
    return this._eventListeners[eventName] || [];
  };

  _proto.emit = function emit(eventName) {
    var handler = this._eventListeners[eventName];

    if (handler) {
      for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
        args[_key - 1] = arguments[_key];
      }

      for (var _i2 = 0; _i2 < handler.length; _i2++) {
        var i = handler[_i2];
        i.apply(void 0, args);
      }
    }
  };

  return EventEmitter;
}();

/**
 * Tag identifying the message, used to match and refine messages.
 */

/**
 * Messages are used to inform states of new events/data which are of interest,
 * these are passed to `StateUpdate` functions of matching states in the
 * state-hierarchy.
 *
 * Extra data can be assigned on the messages, use the `tag` property to
 * differentiate between different messages.
 */

/**
 * A message on its way upwards in the hierarchy.
 */

/**
 * A restricted map of message-key -> subscription-options for a given
 * message-type.
 */

/**
 * Options for a given subcription, the value true means default values for all
 * options.
 */
function isMatchingSubscription(subscriptions, _ref) {
  var tag = _ref.tag;
  return Boolean(subscriptions[tag]);
}

var ANONYMOUS_SOURCE = "$";
var BROADCAST_SOURCE = "@";
var REPLY_SOURCE = "<";

var Supervisor =
/*#__PURE__*/
function (_EventEmitter) {
  _inheritsLoose(Supervisor, _EventEmitter);

  // Explicit constructor results in shorter minified code
  function Supervisor() {
    var _this;

    _this = _EventEmitter.call(this) || this;
    _this._nested = {};
    return _this;
  }
  /**
   * Returns the nested State for the given model and name if it
   * exists, name defaults to model id.
   */


  var _proto = Supervisor.prototype;

  _proto.getState = function getState(m, name) {
    if (process.env.NODE_ENV !== "production") {
      ensureModel(this._getStorage(), m);
    }

    var inst = this._nested[name || m.id];

    if (inst) {
      // TODO: Use path
      debugAssert(inst._name === (name || m.id), "State name '" + inst._name + "' does not match key name '" + (name || m.id));
    }

    return inst;
  }
  /**
   * Attempts to retrieve the nested State for the given model and name,
   * if it does not exist it will be created, name defaults to model id.
   */
  ;

  _proto.createState = function createState(m, params, name) {
    var id = m.id,
        init = m.init;

    if (!name) {
      name = id;
    }

    if (this._nested[name]) {
      // TODO: Maybe add path?
      throw new Error("Duplicate state '" + name + "'");
    }

    var storage = this._getStorage();

    tryAddModel(storage, m);

    var _init = init(params),
        data = _init.data,
        messages = _init.messages;

    debugAssert(this instanceof Storage || this instanceof State, "this is not an instance of Storage or State");
    var instance = new State(id, this, data, name);
    var path = instance.getPath();
    this._nested[name] = instance;
    storage.emit("stateCreated", path, params, data);

    if (messages) {
      instance._init = processEffects(storage, processInstanceMessages(storage, instance._supervisor, messages.map(function (m) {
        return createInflightMessage(storage, path, m);
      })));
    }

    return instance;
  };

  _proto.removeState = function removeState(m, name) {
    var inst = this.getState(m, name);

    if (inst) {
      delete this._nested[name || inst._name];

      this._getStorage().emit("stateRemoved", inst.getPath(), inst._data);
    }
  }
  /**
   * Sends the given message to any matching State or Effect in the
   * state-tree.
   */
  ;

  _proto.sendMessage = function sendMessage(msg, srcName) {
    if (srcName === void 0) {
      srcName = ANONYMOUS_SOURCE;
    }

    var storage = this._getStorage();

    var msgPath = this.getPath().concat([srcName]);
    return processEffects(storage, processInstanceMessages(storage, this, [createInflightMessage(storage, msgPath, msg)]));
  };

  return Supervisor;
}(EventEmitter);
/**
 * Base node in a state-tree, anchors all states and carries all data.
 */


var Storage =
/*#__PURE__*/
function (_Supervisor) {
  _inheritsLoose(Storage, _Supervisor);

  /**
   * Models, used for subscribers, updates and messages.
   */
  // Explicit constructor results in shorter minified code
  function Storage() {
    var _this2;

    _this2 = _Supervisor.call(this) || this;
    _this2._effects = [];
    _this2._running = new Map();
    _this2._defs = {};
    _this2._waiting = [];
    return _this2;
  }
  /**
   * Returns the Storage backing all state in this tree.
   */


  var _proto2 = Storage.prototype;

  _proto2._getStorage = function _getStorage() {
    return this;
  }
  /**
   * Returns the path to this state.
   */
  ;

  _proto2.getPath = function getPath() {
    return [];
  }
  /**
   * Adds the supplied model to the Storage so it can be used when
   * using restoreSnapshot(). Throws if a model with the same id already exist.
   */
  ;

  _proto2.addModel = function addModel(model) {
    if (!tryAddModel(this, model)) {
      throw new Error("Duplicate model '" + model.id + "'.");
    }
  }
  /**
   * Returns the model with the given id, if it exists.
   */
  ;

  _proto2.getModel = function getModel(id) {
    return this._defs[id];
  }
  /**
   * Sends a message to all state-instances currently reachable from this
   * Storage instance.
   */
  ;

  _proto2.broadcastMessage = function broadcastMessage(msg, sourceName) {
    if (sourceName === void 0) {
      sourceName = BROADCAST_SOURCE;
    }

    var inflight = createInflightMessage(this, [sourceName], msg);
    var handled = false;
    /**
     * Broadcasts msg to all state instances with a depth first algo.
     */

    function handleBroadcast(storage, path, nested) {
      var returning = []; // We trust that the user has not been poking around in globals

      /* eslint-disable-next-line guard-for-in */

      for (var key in nested) {
        var instance = nested[key];
        var nestedPath = path.concat([key]);
        returning.push.apply(returning, processMessages(storage, instance, nestedPath, [inflight].concat(handleBroadcast(storage, nestedPath, instance._nested))));
        var i = returning.indexOf(inflight);

        if (i === -1) {
          handled = true;
        } else {
          returning.splice(i, 1);
        }
      }

      return returning;
    }

    var messages = handleBroadcast(this, [], this._nested);

    if (!handled) {
      messages.unshift(inflight);
    }

    return processEffects(this, messages);
  }
  /**
   * Looks up the specified target state and if found the message is processed
   * in the state, and any resulting messages are passed upwards. The reply
   * is always passed to effects at the end.
   */
  ;

  _proto2.replyMessage = function replyMessage(msg, targetState, sourceName) {
    if (sourceName === void 0) {
      sourceName = REPLY_SOURCE;
    }

    var reply = createInflightMessage(this, targetState.concat(sourceName), msg);
    var instance = findState(this, targetState); // If we have an instance, only pass the specific reply message to it and
    // then propagate any resulting messages.

    var inflight = instance ? processInstanceMessages(this, instance._supervisor, processMessages(this, instance, instance.getPath(), [reply])) : [reply];
    return processEffects(this, inflight);
  };

  _proto2.addEffect = function addEffect(eff) {
    this._effects.push(eff);
  };

  _proto2.removeEffect = function removeEffect(eff) {
    var i = this._effects.indexOf(eff);

    if (i !== -1) {
      this._effects.splice(i, 1);
    }
  }
  /**
   * Creates a snapshot of the current state-tree. State-data will not be
   * copied.
   */
  ;

  _proto2.getSnapshot = function getSnapshot() {
    return createSnapshot(this);
  }
  /**
   * Attempts to restore a snapshot. Will throw if required models are missing.
   */
  ;

  _proto2.restoreSnapshot = function restoreSnapshot(snapshot) {
    this.emit("snapshotRestore", snapshot);

    _restoreSnapshot(this, this, snapshot);

    this.emit("snapshotRestored");
  };

  _proto2.wait = function wait() {
    var _this3 = this;

    if (this._running.size === 0) {
      return Promise.resolve();
    }

    return new Promise(function (resolve) {
      _this3._waiting.push(resolve);
    });
  };

  _proto2.runningEffects = function runningEffects() {
    /* eslint-disable-next-line unicorn/prefer-spread */
    return Array.from(this._running.values());
  };

  return Storage;
}(Supervisor);
/**
 * Object representing an instance of a Model.
 */

var State =
/*#__PURE__*/
function (_Supervisor2) {
  _inheritsLoose(State, _Supervisor2);

  /**
   * Matches the Storage _defs collection.
   */

  /**
   * Matches the key used in the supervisor's `_nested` collection.
   */
  function State(id, supervisor, data, name) {
    var _this4;

    _this4 = _Supervisor2.call(this) || this;
    _this4._id = id;
    _this4._name = name;
    _this4._supervisor = supervisor;
    _this4._data = data;
    return _this4;
  }
  /**
   * Returns the data contained in this State.
   */


  var _proto3 = State.prototype;

  _proto3.getData = function getData() {
    return this._data;
  }
  /**
   * Returns the path to this state.
   */
  ;

  _proto3.getPath = function getPath() {
    var path = [];
    var s = this;

    while (s instanceof State) {
      path.unshift(s._name);
      s = s._supervisor;
    }

    return path;
  }
  /**
   * Returns a promise resolving when the state has initialized and its effects
   * have settled. If they already have settled it will return a resolved
   * promise.
   */
  ;

  _proto3.waitInit = function waitInit() {
    if (this._init) {
      return this._init;
    }

    return Promise.resolve(undefined);
  }
  /**
   * Returns the Storage backing all state in this tree.
   */
  ;

  _proto3._getStorage = function _getStorage() {
    var s = this._supervisor;

    while (s instanceof State) {
      s = s._supervisor;
    }

    return s;
  };

  return State;
}(Supervisor);

function _restoreSnapshot(storage, supervisor, snapshot) {
  var newNested = {}; // We trust that the user has not been poking around in globals

  /* eslint-disable-next-line guard-for-in */

  for (var k in snapshot) {
    var _snapshot$k = snapshot[k],
        _id = _snapshot$k.id,
        data = _snapshot$k.data,
        nested = _snapshot$k.nested; // Ensure the model exists when we restore

    getModelById(storage, _id);
    var inst = new State(_id, supervisor, data, k);

    _restoreSnapshot(storage, inst, nested);

    newNested[k] = inst;
  }

  supervisor._nested = newNested;
}
function tryAddModel(storage, model) {
  var id = model.id;

  if (!storage._defs[id]) {
    storage._defs[id] = model;
    return true;
  }

  ensureModel(storage, model);
  return false;
}
function ensureModel(storage, model) {
  var id = model.id;

  if (storage._defs[id] && storage._defs[id] !== model) {
    throw new Error("Model mismatch for '" + id + "'.");
  }
}
function getModelById(storage, id) {
  var spec = storage._defs[id];

  if (!spec) {
    // TODO: Error type
    throw new Error("Missing model for state '" + id + "'.");
  }

  return spec;
}
function createInflightMessage(storage, source, message) {
  storage.emit("messageQueued", message, source);
  return {
    _message: message,
    _source: source,
    _received: false
  };
}
function findState(supervisor, path) {
  for (var _i2 = 0; _i2 < path.length; _i2++) {
    var p = path[_i2];

    if (!supervisor._nested[p]) {
      return null;
    }

    supervisor = supervisor._nested[p];
  }

  return supervisor instanceof State ? supervisor : null;
}
function processInstanceMessages(storage, instance, inflight) {
  var sourcePath = instance.getPath();

  while (instance instanceof State) {
    inflight = processMessages(storage, instance, sourcePath, inflight); // Traverse down one level

    sourcePath = sourcePath.slice(0, -1);
    instance = instance._supervisor;
  }

  return inflight;
}
function processMessages(storage, instance, sourcePath, inflight) {
  var newMessages = [];
  var definition = getModelById(storage, instance._id);
  var update = definition.update;

  for (var _i4 = 0; _i4 < inflight.length; _i4++) {
    var currentInflight = inflight[_i4];
    var m = currentInflight._message;
    var updateRequest = update(instance._data, m);

    if (!updateRequest) {
      // Propagate
      newMessages.push(currentInflight);
      continue;
    }

    storage.emit("messageMatched", m, sourcePath);

    if (updateRequest !== 1) {
      var data = updateRequest.data,
          messages = updateRequest.messages;
      instance._data = data;
      storage.emit("stateNewData", data, sourcePath, m);
      instance.emit("stateNewData", data, sourcePath, m);

      if (messages) {
        for (var _i6 = 0; _i6 < messages.length; _i6++) {
          var _m = messages[_i6];
          newMessages.push(createInflightMessage(storage, sourcePath, _m));
        }
      }
    }
  }

  return newMessages;
}
function processEffects(storage, inflightMsgs) {
  var effects = [];

  outer: for (var _i8 = 0; _i8 < inflightMsgs.length; _i8++) {
    var _ref2 = inflightMsgs[_i8];
    var _message = _ref2._message;
    var _source = _ref2._source;

    for (var _i10 = 0, _storage$_effects2 = storage._effects; _i10 < _storage$_effects2.length; _i10++) {
      var effect = _storage$_effects2[_i10];

      if (!isMatchingSubscription(effect.subscribe, _message)) {
        continue;
      }

      storage.emit("messageMatched", _message, []);
      effects.push(runEffect(storage, effect, _message, _source));
      continue outer;
    }

    storage.emit("unhandledMessage", _message, _source);

    if (storage.listeners("unhandledMessage").length === 0) {
      logUnhandledMessage(_message, _source);
    }
  }

  return Promise.all(effects).then(function () {
    return undefined;
  });
}
/**
 * Runs the supplied effect and attempts to collect any reply-message from it,
 * rejections will cause an `EffectErrorMessage` to be sent as a reply.
 *
 * The resulting promise will always succeed.
 */

function runEffect(storage, _ref3, message, source) {
  var effect = _ref3.effect,
      name = _ref3.name;

  var onSuccess = function onSuccess(msg) {
    return msg ? storage.replyMessage(msg, source, name) : null;
  };

  var onError = function onError(e) {
    return storage.replyMessage({
      tag: EFFECT_ERROR,
      error: e
    }, source, name);
  };

  var promise = new Promise(function (resolve) {
    return resolve(effect(message, source));
  }).then(onSuccess, onError);

  var dropPromise = function dropPromise() {
    storage._running["delete"](promise);

    if (storage._running.size === 0) {
      for (var _i12 = 0, _storage$_waiting2 = storage._waiting; _i12 < _storage$_waiting2.length; _i12++) {
        var r = _storage$_waiting2[_i12];
        r();
      }

      storage._waiting = [];
    }
  };

  storage._running.set(promise, {
    name: name,
    message: message,
    source: source
  });

  return promise.then(dropPromise, dropPromise);
}
function createSnapshot(node) {
  return Object.keys(node._nested).reduce(function (a, key) {
    var nested = node._nested[key];
    a[key] = {
      id: nested._id,
      // We assume it is immutably updated
      data: nested._data,
      nested: createSnapshot(nested)
    };
    return a;
  }, {});
}
function logUnhandledMessage(msg, path) {
  console.error("Unhandled message:", msg, "from [", path.join(", "), "]");
}

exports.EFFECT_ERROR = EFFECT_ERROR;
exports.State = State;
exports.Storage = Storage;
exports.logUnhandledMessage = logUnhandledMessage;
exports.updateAndSend = updateAndSend;
exports.updateData = updateData;
exports.updateNone = updateNone;
//# sourceMappingURL=index.js.map
